/*  $Id: GameEngine.cpp,v 1.7 2009/12/30 18:36:30 sarrazip Exp $
    GameEngine.cpp - Abstract class representing an X11 Game Engine.

    flatzebra - Generic 2D Game Engine library
    Copyright (C) 1999, 2000, 2001 Pierre Sarrazin <http://sarrazip.com/>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
    02111-1307, USA.
*/

#include <flatzebra/GameEngine.h>

#include "font_13x7.xpm"

#include <assert.h>

using namespace std;
using namespace flatzebra;


GameEngine::GameEngine(Couple screenSizeInPixels,
			const string &wmCaption,
			bool fullScreen,
			bool _processActiveEvent) throw(string)
  : theScreenSizeInPixels(screenSizeInPixels),
    theSDLScreen(NULL),
    fixedWidthFontPixmap(NULL),
    theDepth(0),
    processActiveEvent(_processActiveEvent)
{
    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0)
	throw string(SDL_GetError());

    SDL_WM_SetCaption(wmCaption.c_str(), wmCaption.c_str());

    Uint32 flags = SDL_HWSURFACE | SDL_ANYFORMAT;
    if (fullScreen)
	flags |= SDL_FULLSCREEN;

    theDepth = SDL_VideoModeOK(
			screenSizeInPixels.x, screenSizeInPixels.y,
			32, flags);
    if (theDepth <= 0)
    {
	SDL_Quit();
	throw string("video mode not available");
    }

    theSDLScreen = SDL_SetVideoMode(
    			screenSizeInPixels.x, screenSizeInPixels.y,
			theDepth, flags);
    if (theSDLScreen == NULL)
    {
	string errMsg = SDL_GetError();
	SDL_Quit();
	throw errMsg;
    }

    // Deselect unused event types:
    for (int i = SDL_NOEVENT; i < SDL_NUMEVENTS; i++)
    {
	switch (i)
	{
	    case SDL_KEYDOWN:
	    case SDL_KEYUP:
	    case SDL_QUIT:
		break;
	    case SDL_ACTIVEEVENT:
		if (processActiveEvent)
		    break;
	    default:
		SDL_EventState((Uint8) i, SDL_IGNORE);
	}
    }

    // Prepare a fixed width font pixmap:
    try
    {
	Couple dummy;
	loadPixmap(font_13x7_xpm, fixedWidthFontPixmap, dummy);
	assert(fixedWidthFontPixmap != NULL);
    }
    catch (PixmapLoadError)
    {
	throw string("Could not load fixed width font pixmap");
    }

    // Hide the mouse pointer when in full-screen mode:
    if (fullScreen)
	SDL_ShowCursor(SDL_DISABLE);
}


GameEngine::~GameEngine()
{
    SDL_FreeSurface(fixedWidthFontPixmap);
    SDL_Quit();
}


void GameEngine::run(int millisecondsPerFrame)
{
    for (;;)
    {
	Uint32 lastTime = SDL_GetTicks();

	SDL_Event event;
	while (SDL_PollEvent(&event))
	{
	    if (event.type == SDL_KEYDOWN)
		processKey(event.key.keysym.sym, true);   // virtual function
	    else if (event.type == SDL_KEYUP)
		processKey(event.key.keysym.sym, false);  // virtual function
	    else if (processActiveEvent && event.type == SDL_ACTIVEEVENT)
	    {
		// If app is now inactive (iconified or lost focus),
		// go wait for reactivation, i.e., stop calling tick().
		// This avoids using the CPU while the app is not active.
		//
		if (event.active.gain == 0 && !waitForReactivation())
		    return;  // wait ended with quit or with wait error
	    }
	    else if (event.type == SDL_QUIT)
		return;
	}

	if (!tick())  // virtual function
	    return;

	SDL_Flip(theSDLScreen);

	// Pause for the rest of the current animation frame.
	Uint32 limit = lastTime + millisecondsPerFrame;
	Uint32 delay = limit - SDL_GetTicks();
	if (delay <= (Uint32) millisecondsPerFrame)
	    SDL_Delay(delay);
    }
}


bool
GameEngine::waitForReactivation()
{
    // Virtual call: notify derived class that app is now inactive.
    //
    processActivation(false);

    // Allow effect of drawing commands made by processActivation() to appear.
    //
    SDL_Flip(theSDLScreen);

    // Sleep on SDL event loop until reactivation.
    //
    SDL_Event event;
    while (SDL_WaitEvent(&event))
    {
	// Ignore events other than quitting and reactivation.
	//
	switch (event.type)
	{
	    case SDL_QUIT:
		processActivation(true);  // notify derived class that app is now active
		return false;  // ask caller to quit
	    case SDL_ACTIVEEVENT:
		if (event.active.gain != 0)  // if app now active
		{
		    processActivation(true);
		    return true;  // ask caller to continue
		}
		break;
	}
    }
    return false;  // unexpected wait error: quit
}


/*virtual*/
void
GameEngine::processActivation(bool /*appActive*/)
{
}


void
GameEngine::loadPixmap(const char **xpmData, PixmapArray &pa, size_t index) const
						throw(PixmapLoadError)
{
    // Masks are not be relevant with SDL.

    SDL_Surface *pixmap;
    Couple size;
    loadPixmap(xpmData, pixmap, size);
    pa.setArrayElement(index, pixmap);
    pa.setImageSize(size);
}


void
GameEngine::loadPixmap(const char **xpmData,
		SDL_Surface *&pixmap,
		Couple &pixmapSize) const throw(PixmapLoadError)
{
    pixmapSize.zero();

    if (xpmData == NULL || xpmData[0] == NULL)
	throw PixmapLoadError(PixmapLoadError::INVALID_ARGS, NULL);

    pixmap = IMG_ReadXPMFromArray(const_cast<char **>(xpmData));
    if (pixmap == NULL)
	throw PixmapLoadError(PixmapLoadError::UNKNOWN, NULL);

    pixmapSize.x = pixmap->w;
    pixmapSize.y = pixmap->h;
}


void
GameEngine::writeString(const char *s, Couple pos, SDL_Surface *surface)
{
    assert(fixedWidthFontPixmap != NULL);
    if (s == NULL)
	return;
    if (surface == NULL)
	surface = theSDLScreen;
    Couple fontDim = getFontDimensions();
    SDL_Rect dest = { Sint16(pos.x), Sint16(pos.y), Uint16(fontDim.x), Uint16(fontDim.y) };
    for (size_t i = 0; s[i] != '\0'; i++, dest.x += Sint16(fontDim.x))
    {
	unsigned char c = (unsigned char) s[i];
	if (c < 32 || (c >= 127 && c <= 160))  // if ctrl char or undef char
	    c = 32;  // replace by space

	/*  Compute the subrectangle of fixedWidthFontPixmap that
	    contains the character:
	*/
	int x = (c % 16) * fontDim.x;
	int y = (c - 32) / 16;
	if (y >= 8)
	    y -= 2;
	y *= fontDim.y;

	SDL_Rect src  = { Sint16(x), Sint16(y), Uint16(fontDim.x), Uint16(fontDim.y) };
	SDL_BlitSurface(fixedWidthFontPixmap, &src, surface, &dest);
    }
}


void
GameEngine::writeStringCentered(const char *s, Couple pos, SDL_Surface *surface)
{
    Couple fontDim = getFontDimensions();
    Couple stringSizeInPixels(strlen(s) * fontDim.x, fontDim.y);
    writeString(s, pos - stringSizeInPixels / 2, surface);
}


void
GameEngine::writeStringXCentered(
			const char *s, Couple pos, SDL_Surface *surface)
{
    Couple fontDim = getFontDimensions();
    int stringWidthInPixels = strlen(s) * fontDim.x;
    Couple adjustedPos(pos.x - stringWidthInPixels / 2, pos.y);
    writeString(s, adjustedPos, surface);
}


void
GameEngine::writeStringRightJustified(
			const char *s, Couple pos, SDL_Surface *surface)
{
    Couple fontDim = getFontDimensions();
    int stringWidthInPixels = strlen(s) * fontDim.x;
    Couple adjustedPos(pos.x - stringWidthInPixels, pos.y);
    writeString(s, adjustedPos, surface);
}
